from ..annotations import DictStrAny, DictStrStr


def format_version(schema: str, parts: DictStrStr, context: DictStrAny) -> str:
    for part, template in parts.items():
        schema = schema.replace(part, template)
    return schema.format(**context)



import argparse
import os
import sys
from typing import Optional

from .arguments import add_path_argument
from .commands import (
    run_post_bump_hook,
    update_changelog_file,
    update_version_files,
)
from .output import echo_value, EMPTY, github_actions_output
from .. import __app__, __version__
from ..annotations import Argv
from ..changelog import ChangeLog
from ..configs import ProjectConfig, UpdateConfig
from ..constants import INITIAL_PRE_RELEASE_COMMIT, INITIAL_RELEASE_COMMIT
from ..enums import ChangeLogTypeEnum
from ..git import Git
from ..versions import Version


def create_update_config(
    changelog: ChangeLog, is_pre_release: bool
) -> UpdateConfig:
    kwargs = {
        "is_breaking_change": False,
        "is_minor_change": False,
        "is_micro_change": False,
        "is_pre_release": is_pre_release,
    }

    if changelog.has_breaking_change:
        kwargs["is_breaking_change"] = True
    elif changelog.has_minor_change:
        kwargs["is_minor_change"] = True
    else:
        kwargs["is_micro_change"] = True

    return UpdateConfig(**kwargs)


def parse_args(argv: Argv) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog=__app__,
        description=(
            "Manage changelog and bump project version number using "
            "conventional commits from latest git tag. Support Python, "
            "JavaScript projects. Understand CalVer, SemVer version schemas. "
            "Designed to run at GitHub Actions."
        ),
    )
    parser.add_argument(
        "-v", "--version", action="version", version=__version__
    )
    add_path_argument(parser)
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        default=False,
        dest="is_dry_run",
        help="Only show updates, but do not apply them.",
    )
    parser.add_argument(
        "--ci",
        action="store_true",
        default=os.getenv("CI") is not None,
        dest="is_ci",
        help=f"Run {__app__} in CI mode.",
    )
    parser.add_argument(
        "--pre",
        action="store_true",
        default=False,
        dest="is_pre_release",
        help="Pre-release change. By default: False",
    )
    return parser.parse_args(argv)


def main(argv: Argv = None) -> int:
    args = parse_args(argv or sys.argv[1:])

    # Initialize project config
    project_config = ProjectConfig.from_path(args.path)

    # Read latest git tag and parse current version
    git = Git(path=project_config.path)

    current_tag = git.retrieve_last_tag_or_none()
    echo_value(
        "Current tag: ",
        current_tag or EMPTY,
        is_ci=args.is_ci,
        ci_name="current_tag",
    )

    current_version: Optional[Version] = None
    if current_tag is not None:
        current_version = Version.from_tag(current_tag, config=project_config)

    echo_value(
        "Current version: ",
        (
            current_version.format(config=project_config)
            if current_version
            else EMPTY
        ),
        is_ci=args.is_ci,
        ci_name="current_version",
    )

    # Read commits from last tag
    if current_tag is not None and current_version is not None:
        try:
            git_commits = git.list_commits(current_tag)
            if not git_commits and current_version.pre_release is None:
                raise ValueError("No commits found after latest tag")
        except ValueError:
            print(
                f"ERROR: No commits found after: {current_tag!r}. Exit...",
                file=sys.stderr,
            )
            return 1

        # Create changelog using commits from last tag
        changelog = ChangeLog.from_git_commits(git_commits)
    # Create initial changelog
    else:
        changelog = ChangeLog.from_git_commits(
            (
                INITIAL_PRE_RELEASE_COMMIT
                if args.is_pre_release
                else INITIAL_RELEASE_COMMIT,
            ),
        )

    git_changelog = changelog.format(
        ChangeLogTypeEnum.git_commit,
        project_config.changelog_format_type_git,
    )
    echo_value(
        "\nChangeLog\n\n",
        git_changelog,
        is_ci=args.is_ci,
        ci_name="changelog",
    )

    # Supply update config and guess next version
    update_config = create_update_config(changelog, args.is_pre_release)

    if current_version is not None:
        next_version = current_version.update(update_config)
    else:
        next_version = Version.guess_initial_version(
            config=project_config, is_pre_release=args.is_pre_release
        )

    next_version_str = next_version.format(config=project_config)
    echo_value(
        "\nNext version: ",
        next_version_str,
        is_ci=args.is_ci,
        ci_name="next_version",
    )

    # Applying changes to version files
    if not args.is_ci and not args.is_dry_run:
        if (
            input("Are you sure to update version files? [y/N] ").lower()
            != "y"
        ):
            print("OK! OK! Exit...")
            return 0

    update_version_files(
        project_config,
        current_version,
        next_version,
        is_dry_run=args.is_dry_run,
    )

    # Run post-bump hook
    run_post_bump_hook(project_config, is_dry_run=args.is_dry_run)

    # Update changelog
    update_changelog_file(
        project_config, next_version, changelog, is_dry_run=args.is_dry_run
    )

    # Supply necessary CI output
    if args.is_ci:
        github_actions_output(
            "next_tag",
            project_config.tag_format.format(version=next_version_str),
        )
        github_actions_output(
            "next_tag_message",
            "\n\n".join(
                (
                    project_config.tag_subject_format.format(
                        version=next_version_str
                    ),
                    git_changelog,
                )
            ),
        )
        github_actions_output(
            "pr_branch",
            project_config.pr_branch_format.format(version=next_version_str),
        )
        github_actions_output(
            "pr_title",
            project_config.pr_title_format.format(version=next_version_str),
        )

    print("All OK!")
    return 0



import json
from contextlib import suppress
from typing import cast, Optional, Type, Union

import attr
import toml

from . import calver, pre_release, semver
from .calver import CalVer
from .exceptions import VersionError, VersionParseError
from .parsing import parse_version
from .pre_release import PreRelease
from .semver import SemVer
from ..annotations import DictStrStr
from ..configs import ProjectConfig, UpdateConfig
from ..enums import ProjectTypeEnum, VersionTypeEnum
from ..regexps import to_regexp


CalOrSemVer = Union[CalVer, SemVer]


@attr.dataclass(frozen=True, slots=True)
class Version:
    version: CalOrSemVer
    pre_release: Optional[PreRelease] = None

    @classmethod
    def from_tag(cls, value: str, *, config: ProjectConfig) -> "Version":
        return cls.parse(
            guess_version_from_tag(value, tag_format=config.tag_format),
            config=config,
        )

    def format(self, *, config: ProjectConfig) -> str:  # noqa: A003
        if self.pre_release:
            return (
                f"{self.version.format()}"
                f"{self.pre_release.format(project_type=config.project_type)}"
            )
        return self.version.format()

    @classmethod
    def guess_initial_version(
        cls, *, config: ProjectConfig, is_pre_release: bool
    ) -> "Version":
        version = guess_initial_version(config)
        if version.pre_release is None and is_pre_release:
            return attr.evolve(version, pre_release=PreRelease())
        return version

    @classmethod
    def parse(cls, value: str, *, config: ProjectConfig) -> "Version":
        schema = config.version_schema

        version_cls: Type[CalOrSemVer]
        full_schema_parts: DictStrStr
        if config.version_type == VersionTypeEnum.semver:
            version_cls = semver.SemVer
            full_schema_parts = semver.SCHEMA_PARTS_PARSING
        else:
            version_cls = calver.CalVer
            full_schema_parts = calver.SCHEMA_PARTS_PARSING

        with suppress(VersionError):
            return cls(version=version_cls.parse(value, schema=schema))

        full_schema = (
            f"{schema}{pre_release.SCHEMA_MAPPING[config.project_type]}"
        )
        full_schema_parts.update(pre_release.SCHEMA_PARTS_PARSING)

        maybe_parsed = parse_version(full_schema, full_schema_parts, value)
        if maybe_parsed:
            return cls(
                version=version_cls.from_parsed_dict(
                    maybe_parsed, schema=schema
                ),
                pre_release=pre_release.PreRelease.from_parsed_dict(
                    maybe_parsed, project_type=config.project_type
                ),
            )

        raise VersionParseError(schema, value)

    def update(self, config: UpdateConfig) -> "Version":
        if self.pre_release:
            return Version(
                version=self.version,
                pre_release=self.pre_release.update(config),
            )

        if config.is_pre_release:
            return Version(
                version=self.version.update(
                    attr.evolve(config, is_pre_release=False)
                ),
                pre_release=PreRelease(),
            )

        return Version(version=self.version.update(config))


def find_project_version(config: ProjectConfig) -> Optional[str]:
    if config.project_type == ProjectTypeEnum.javascript:
        package_json_path = config.path / "package.json"
        if package_json_path.exists():
            try:
                return cast(
                    str, json.loads(package_json_path.read_text())["version"]
                )
            except (KeyError, ValueError):
                ...
    else:
        pyproject_toml_path = config.path / "pyproject.toml"
        if pyproject_toml_path.exists():
            try:
                return cast(
                    str,
                    toml.loads(pyproject_toml_path.read_text())["tool"][
                        "poetry"
                    ]["version"],
                )
            except (KeyError, ValueError):
                ...

    return None


def guess_initial_version(config: ProjectConfig) -> Version:
    maybe_version_str = find_project_version(config)
    if maybe_version_str:
        return Version.parse(maybe_version_str, config=config)

    if config.version_type == VersionTypeEnum.semver:
        return Version(version=SemVer.initial())
    return Version(version=CalVer.initial(schema=config.version_schema))


def guess_version_from_tag(value: str, *, tag_format: str) -> str:
    matched = to_regexp(tag_format).match(value)
    if matched:
        return matched.groupdict()["version"]
    raise ValueError("Tag value does not match given tag format")



from contextlib import suppress
from enum import Enum, unique
from typing import DefaultDict, Optional

import attr

from .formatting import format_version
from .parsing import parse_version
from ..annotations import DictStrAny, DictStrStr
from ..configs import UpdateConfig
from ..enums import ProjectTypeEnum


SCHEMA_MAPPING: DefaultDict[ProjectTypeEnum, str] = DefaultDict(
    lambda: "-TYPE.NUMBER"
)
SCHEMA_MAPPING[ProjectTypeEnum.python] = "TYPENUMBER"

SCHEMA_PARTS_FORMATTING = {
    "TYPE": "{type}",
    "NUMBER": "{number}",
}
SCHEMA_PARTS_PARSING = {
    "TYPE": r"(?P<type>.+)",
    "NUMBER": r"(?P<number>\d+)",
}


@unique
class PreReleaseTypeEnum(Enum):
    alpha = "alpha"
    beta = "beta"
    rc = "rc"


NEXT_PRE_RELEASE_TYPE = {
    PreReleaseTypeEnum.alpha: PreReleaseTypeEnum.beta,
    PreReleaseTypeEnum.beta: PreReleaseTypeEnum.rc,
    PreReleaseTypeEnum.rc: PreReleaseTypeEnum.rc,
}

PRE_RELEASE_TYPE_MAPPING = {
    ProjectTypeEnum.python: {
        PreReleaseTypeEnum.alpha: "a",
        PreReleaseTypeEnum.beta: "b",
        PreReleaseTypeEnum.rc: "rc",
    },
}


@attr.dataclass(frozen=True, slots=True)
class PreRelease:
    pre_release_type: PreReleaseTypeEnum = PreReleaseTypeEnum.alpha
    number: int = 0

    def format(self, *, project_type: ProjectTypeEnum) -> str:  # noqa: A003
        schema = SCHEMA_MAPPING[project_type]

        context: DictStrAny = {"number": self.number}
        maybe_type_mapping = PRE_RELEASE_TYPE_MAPPING.get(project_type)
        if maybe_type_mapping:
            context["type"] = maybe_type_mapping[self.pre_release_type]
        else:
            context["type"] = self.pre_release_type.value

        return format_version(schema, SCHEMA_PARTS_FORMATTING, context)

    @classmethod
    def from_parsed_dict(
        cls, parsed_dict: DictStrStr, *, project_type: ProjectTypeEnum
    ) -> "PreRelease":
        return cls(
            pre_release_type=guess_pre_release_type(
                parsed_dict["type"], project_type=project_type
            ),
            number=int(parsed_dict["number"]),
        )

    @classmethod
    def parse(
        cls, value: str, *, project_type: ProjectTypeEnum
    ) -> "PreRelease":
        schema = SCHEMA_MAPPING[project_type]

        maybe_parsed = parse_version(schema, SCHEMA_PARTS_PARSING, value)
        if maybe_parsed:
            with suppress(KeyError):
                return cls.from_parsed_dict(
                    maybe_parsed, project_type=project_type
                )

        raise ValueError(
            "Invalid pre-release value, which do not match any supported "
            "project type"
        )

    def update(self, config: UpdateConfig) -> Optional["PreRelease"]:
        if config.is_pre_release is False:
            return None

        if config.is_breaking_change:
            next_type = NEXT_PRE_RELEASE_TYPE[self.pre_release_type]
            if next_type != self.pre_release_type:
                return PreRelease(pre_release_type=next_type, number=0)
            return PreRelease(
                pre_release_type=next_type, number=self.number + 1
            )

        return PreRelease(
            pre_release_type=self.pre_release_type, number=self.number + 1
        )


def guess_pre_release_type(
    value: str, *, project_type: ProjectTypeEnum
) -> PreReleaseTypeEnum:
    maybe_type_mapping = PRE_RELEASE_TYPE_MAPPING.get(project_type)
    if maybe_type_mapping:
        return {
            item: pre_release_type
            for pre_release_type, item in maybe_type_mapping.items()
        }[value]
    return PreReleaseTypeEnum[value]



class VersionError(Exception):
    """Base version error"""


class VersionParseError(VersionError, ValueError):
    """Unable to parse version"""

    schema: str
    value: str

    def __init__(self, schema: str, value: str) -> None:
        super().__init__(
            "Unable to parse version using given schema & value. Expected "
            f"schema: {schema}, value: {value}"
        )

        self.schema = schema
        self.value = value



from enum import Enum, unique


@unique
class ChangeLogTypeEnum(Enum):
    changelog_file = "changelog_file"
    git_commit = "git_commit"


@unique
class FormatTypeEnum(Enum):
    markdown = "markdown"
    rst = "rst"


@unique
class ProjectTypeEnum(Enum):
    python = "python"
    javascript = "javascript"


@unique
class VersionTypeEnum(Enum):
    calver = "calver"
    semver = "semver"



from pathlib import Path
from typing import Optional, Tuple

import attr
import toml

from . import __app__
from .annotations import DictStrAny
from .constants import (
    CHANGELOG_LOWER,
    DEFAULT_CHANGELOG_FILE_INCLUDE_DATE,
    DEFAULT_CHANGELOG_FORMAT_TYPE_FILE,
    DEFAULT_CHANGELOG_FORMAT_TYPE_GIT,
    DEFAULT_PR_BRANCH_FORMAT,
    DEFAULT_PR_TITLE_FORMAT,
    DEFAULT_PROJECT_TYPE,
    DEFAULT_SEMVER_SCHEMA,
    DEFAULT_TAG_FORMAT,
    DEFAULT_TAG_SUBJECT_FORMAT,
    DEFAULT_VERSION_SCHEMA,
    DEFAULT_VERSION_TYPE,
)
from .enums import FormatTypeEnum, ProjectTypeEnum, VersionTypeEnum


@attr.dataclass(frozen=True, slots=True)
class ProjectConfig:
    path: Path = attr.Factory(Path.cwd)

    project_type: ProjectTypeEnum = DEFAULT_PROJECT_TYPE

    version_type: VersionTypeEnum = DEFAULT_VERSION_TYPE
    version_schema: str = DEFAULT_VERSION_SCHEMA
    version_files: Tuple[str, ...] = attr.Factory(tuple)

    tag_format: str = DEFAULT_TAG_FORMAT
    tag_subject_format: str = DEFAULT_TAG_SUBJECT_FORMAT
    pr_branch_format: str = DEFAULT_PR_BRANCH_FORMAT
    pr_title_format: str = DEFAULT_PR_TITLE_FORMAT

    changelog_format_type_file: FormatTypeEnum = (
        DEFAULT_CHANGELOG_FORMAT_TYPE_FILE
    )
    changelog_format_type_git: FormatTypeEnum = (
        DEFAULT_CHANGELOG_FORMAT_TYPE_GIT
    )
    changelog_file_include_date: bool = DEFAULT_CHANGELOG_FILE_INCLUDE_DATE

    post_bump_hook: Optional[str] = None

    def __attrs_post_init__(self) -> None:
        if self.version_type == VersionTypeEnum.semver:
            object.__setattr__(self, "version_schema", DEFAULT_SEMVER_SCHEMA)

    @classmethod
    def from_path(cls, path: Path) -> "ProjectConfig":
        if not path.is_dir() or not path.exists():
            raise ValueError(f"Project path does not exist: {path!r}")

        config_data: DictStrAny = {}

        maybe_loaded = load_project_config_data(path)
        if maybe_loaded:
            config_data = maybe_loaded[1]

        maybe_include_date = config_data.get("changelog_file_include_date")

        return cls(
            path=path,
            project_type=guess_project_type(
                config_data.get("project_type"), path=path
            ),
            version_type=guess_version_type(config_data.get("version_type")),
            version_schema=(
                config_data.get("version_schema") or DEFAULT_VERSION_SCHEMA
            ),
            version_files=tuple(config_data.get("version_files") or ()),
            tag_format=config_data.get("tag_format") or DEFAULT_TAG_FORMAT,
            tag_subject_format=(
                config_data.get("tag_subject_format")
                or DEFAULT_TAG_SUBJECT_FORMAT
            ),
            pr_branch_format=(
                config_data.get("pr_branch_format") or DEFAULT_PR_BRANCH_FORMAT
            ),
            pr_title_format=(
                config_data.get("pr_title_format") or DEFAULT_PR_TITLE_FORMAT
            ),
            changelog_format_type_file=guess_changelog_format_type_file(
                config_data.get("changelog_format_type_file"), path
            ),
            changelog_format_type_git=guess_changelog_format_type_git(
                config_data.get("changelog_format_type_git")
            ),
            changelog_file_include_date=(
                maybe_include_date
                if maybe_include_date is not None
                else DEFAULT_CHANGELOG_FILE_INCLUDE_DATE
            ),
            post_bump_hook=config_data.get("post_bump_hook"),
        )


@attr.dataclass(frozen=True, slots=True)
class UpdateConfig:
    is_breaking_change: bool = False
    is_minor_change: bool = False
    is_micro_change: bool = True
    is_pre_release: bool = False

    def __attrs_post_init__(self) -> None:
        values = sorted(
            [
                self.is_breaking_change,
                self.is_minor_change,
                self.is_micro_change,
            ]
        )
        if values != [False, False, True]:
            raise ValueError(
                "Update config allow to be initialized only with one "
                "breaking, minor, or micro change at a time."
            )


def guess_changelog_format_type_file(
    value: Optional[str], path: Path
) -> FormatTypeEnum:
    if value:
        return FormatTypeEnum[value]

    if find_changelog_file(path, "*.md") is not None:
        return FormatTypeEnum.markdown

    if find_changelog_file(path, "*.rst") is not None:
        return FormatTypeEnum.rst

    return DEFAULT_CHANGELOG_FORMAT_TYPE_FILE


def find_changelog_file(path: Path, pattern: str) -> Optional[Path]:
    for item in path.glob(pattern):
        if item.stem.lower() == CHANGELOG_LOWER:
            return item
    return None


def guess_changelog_format_type_git(value: Optional[str]) -> FormatTypeEnum:
    if value:
        return FormatTypeEnum[value]
    return DEFAULT_CHANGELOG_FORMAT_TYPE_GIT


def guess_project_type(value: Optional[str], path: Path) -> ProjectTypeEnum:
    if value:
        return ProjectTypeEnum[value]

    if (path / "package.json").exists():
        return ProjectTypeEnum.javascript

    return DEFAULT_PROJECT_TYPE


def guess_version_type(value: Optional[str]) -> VersionTypeEnum:
    if value:
        return VersionTypeEnum[value]
    return DEFAULT_VERSION_TYPE


def load_project_config_data(path: Path) -> Optional[Tuple[Path, DictStrAny]]:
    for item in (f".{__app__}.toml", "pyproject.toml"):
        maybe_config_path = path / item
        if not maybe_config_path.exists():
            continue

        data = toml.loads(maybe_config_path.read_text())
        return (maybe_config_path, data.get("tool", {}).get(__app__, {}) or {})

    return None



import re
from typing import Optional, Pattern

from ..annotations import DictStrStr
from ..regexps import ensure_regexp_dots


def build_schema_regexp(schema: str, parts: DictStrStr) -> Pattern[str]:
    schema = ensure_regexp_dots(schema)
    for part, regexp in parts.items():
        schema = schema.replace(part, regexp)
    return re.compile(rf"^{schema}$")


def parse_version(
    schema: str, parts: DictStrStr, value: str
) -> Optional[DictStrStr]:
    maybe_matched = build_schema_regexp(schema, parts).match(value)
    if maybe_matched:
        return maybe_matched.groupdict()
    return None



import re
from typing import Pattern


VAR_RE = re.compile(r"\{(?P<var>[^\{]+)\}")


def ensure_regexp_dots(value: str) -> str:
    return value.replace(".", r"\.")


def to_regexp(value: str) -> Pattern[str]:
    value = ensure_regexp_dots(value)
    for item in VAR_RE.findall(value):
        value = value.replace(f"{{{item}}}", rf"(?P<{item}>.+)")
    return re.compile(rf"^{value}$")
